//muroar.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2009
 *
 *  This file is part of µRoar,
 *  a minimum library to access a RoarAudio Sound Server.
 *
 *  µRoar is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  µRoar is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with µRoar.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "muroar.h"
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h> /* for snprintf() */

#ifdef __WIN32
#include <windows.h>
#include <winsock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#if defined(PF_UNIX)
#include <sys/un.h>
#endif
#include <netdb.h>
#endif

#define MUROAR_IOBUF  128

#define _OK 254

#define _MU_STR "\265"

#ifdef __WIN32
#define _CLOSE(x) closesocket((x))
#else
#define _CLOSE(x) close((x))
#endif

#ifdef __WIN32
static void muroar_init_win32 (void) {
 static int inited = 0;
 WSADATA wsadata;

 if ( !inited ) {
  WSAStartup(MAKEWORD(1,1) , &wsadata);
  inited++;
 }
}
#endif

static int muroar_open_socket (char * server) {
 struct hostent     * he;
 struct sockaddr_in   in;
#if !defined(__WIN32) && defined(PF_UNIX)
 struct sockaddr_un   un;
#endif
 int fh = -1;

#ifdef __WIN32
 muroar_init_win32();
#endif

 if ( *server == '/' ) {
#if !defined(__WIN32) && defined(PF_UNIX)
  if ( (fh = socket(PF_UNIX, SOCK_STREAM, 0)) == -1 )
   return -1;

  un.sun_family = PF_UNIX;
  strncpy(un.sun_path, server, sizeof(un.sun_path) - 1);

  if ( connect(fh, (struct sockaddr *)&un, sizeof(struct sockaddr_un)) == -1 ) {
   _CLOSE(fh);
   return -1;
  }

  return fh;
#else
  return -1;
#endif
 }

 if ( (he = gethostbyname(server)) == NULL ) {
  return -1;
 }

 if ( (fh = socket(PF_INET, SOCK_STREAM, 0)) == -1 )
  return -1;

 memcpy((struct in_addr *)&in.sin_addr, he->h_addr, sizeof(struct in_addr));

 in.sin_family = AF_INET;
 in.sin_port   = htons(MUROAR_PORT);

 if ( connect(fh, (const struct sockaddr *)&in, sizeof(in)) == -1 ) {
  _CLOSE(fh);
  return -1;
 }

 return fh;
}

int muroar_connect(char * server, char * name) {
 char   useraddr[80] = "/invalid";
 char * addr[] = {useraddr, "/tmp/roar", "localhost", NULL};
 unsigned char buf[MUROAR_IOBUF];
 uint16_t tmpu16;
 int fh = -1;
 int i;

 if ( server != NULL && *server == 0 )
  server = NULL;

 if ( server == NULL )
  server = getenv("ROAR_SERVER");

 if ( name == NULL || *name == 0 )
  name = _MU_STR "Roar Client";

 // ROAR_CMD_IDENTIFY:
 memset(buf, 0, sizeof(buf));
 buf[1] = MUROAR_CMD_IDENTIFY;

 tmpu16 = strlen(name) + 5;

 if ( tmpu16 > MUROAR_IOBUF )
  return -1;

 buf[8] = (tmpu16 & 0xFF00) >> 8;
 buf[9] = (tmpu16 & 0x00FF);

 if ( server != NULL ) {
  if ( (fh = muroar_open_socket(server)) == -1 )
   return -1;
 } else {
  snprintf(useraddr, sizeof(useraddr), "/%s/.roar", getenv("HOME"));
  useraddr[sizeof(useraddr)-1] = 0;

  for (i = 0; fh == -1 && addr[i] != NULL; i++) {
   fh = muroar_open_socket(addr[i]);
  }

  if ( fh == -1 )
   return -1;
 }

 if ( send(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 buf[0] = 1;
 *(uint32_t*)(&(buf[1])) = htonl(getpid());

 // sizes are allready checked.
 strcpy((char*)&(buf[5]), name);

 if ( send(fh, buf, tmpu16, 0) != tmpu16 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( recv(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( buf[1] != _OK ) {
  _CLOSE(fh);
  return -1;
 }

 // ROAR_CMD_AUTH:
 memset(buf, 0, 10);
 buf[1] = MUROAR_CMD_AUTH;

 if ( send(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( recv(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( buf[1] != _OK ) {
  _CLOSE(fh);
  return -1;
 }

 return fh;
}

int muroar_close(int fh) {
 if ( fh == -1 )
  return -1;

 _CLOSE(fh);

 return 0;
}

int muroar_quit   (int fh) {
 char quit[] = "\0\6\0\0\0\0\0\0\0\0";
 int ret = 0;

 if ( fh == -1 )
  return -1;

 if ( send(fh, quit, 10, 0) != 10 )
  ret = -1;

 // read in case the server response
 // ignore errors as the server do not nessessery
 // response to our request.
 recv(fh, quit, 10, 0);

 if ( muroar_close(fh) == -1 )
  ret = -1;

 return ret;
}

int muroar_beep   (int fh) {
 unsigned char beep[] = "\0\30\0\0\0\0\0\0\0\0";

 if ( fh == -1 )
  return -1;

 if ( send(fh, beep, 10, 0) != 10 )
  return -1;

 if ( recv(fh, beep, 10, 0) != 10 )
  return -1;

 if ( beep[1] != _OK )
  return -1;

 return 0;
}

int muroar_stream(int fh, int dir, int * stream, int codec, int rate, int channels, int bits) {
 unsigned char buf[24];
 uint16_t sid;
 uint32_t * data = (uint32_t*)buf;
 int i;

 if ( fh == -1 )
  return -1;

 // ROAR_CMD_NEW_STREAM  = 3:
 memset(buf, 0, 10);
 buf[1] = MUROAR_CMD_NEW_STREAM;
 buf[9] = 24;

 if ( send(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 data[0] = dir;
 data[1] = -1;
 data[2] = rate;
 data[3] = bits;
 data[4] = channels;
 data[5] = codec;

 for (i = 0; i < 6; i++)
  data[i] = htonl(data[i]);

 if ( send(fh, buf, 24, 0) != 24 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( recv(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( buf[1] != _OK ) {
  _CLOSE(fh);
  return -1;
 }

 sid = (buf[2] << 8) | buf[3];

 // ROAR_CMD_EXEC_STREAM = 5:

 memset(buf, 0, 10);
 buf[1] = MUROAR_CMD_EXEC_STREAM;
 buf[2] = (sid & 0xFF00) >> 8;
 buf[3] = (sid & 0x00FF);

 if ( send(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( recv(fh, buf, 10, 0) != 10 ) {
  _CLOSE(fh);
  return -1;
 }

 if ( stream != NULL )
  *stream = sid;

 return fh;
}

//ll
